/*

This benchmark app will test various performance marks on NetBurner products.
The various tests can be run individually or together.  To enable/disable the 
tests, comment out the following enable defines:
*/
#define IRQ_TEST
#define OS_TEST
#define UDP_TEST
#define TCP_TEST
#define SRAM_INFO
//#define RamTestEnable

// CHOOSE ONLY 1
#define UDP_SEND_KEEP
//#define UDP_SEND_NEW
//#define UDP_SEND_SOCKET

/*
The best results will be  seen when running the individual tests alone due to
 resource availability.  For example,  if the TCP test will have lower performance 
 when used at the same time as the RAM test since the RAM test consumes some 
onchip SRAM.  The test reluts will also vary depending on which parts of the 
system you have located in SRAM.  This is adjusted in constants.h and 
documented here:
http://www.netburner.com/downloads/nndk/SRAM-Performance.pdf
One example of this is the idle task stack.  When located in SRAM there 
will be about a 10% gain in UDP performance but a 10% drop in TCP
performance.  This application should help you decide how to adjust the
system SRAM usage to best fit your application.


*/

#include "predef.h"
#include <stdio.h>
#include <ctype.h>
#include <startnet.h>
#include <autoupdate.h>
#include <dhcpclient.h>
#include <smarttrap.h>
#include <taskmon.h>
#include <NetworkDebug.h>
#include <cfinter.h>
#include <ip.h>
#include <udp.h>
#include <tcp.h>
#include <bsp.h>
#include <buffers.h>
#include <ethernet.h>
#include "StopWatch.h"
#include <constants.h>
#include <stdlib.h>
#include <string.h>

#ifdef MOD5270B
#ifndef MOD5270
#define MOD5270
#endif
#endif

#include <sim.h>

#ifdef MCF53017
#define eport eport[0]
#endif

const char * AppName="NB Benchmark";

extern "C" {
	void NMI_C_Part();
void NMI_ASM_Part(); // The part of the ISR written in Assembly
#if (defined MCF5272 || defined MCF5208 )
#elif ( defined MCF5282 || defined MCF5234 || defined MCF53017 )
void SetIntc( int intc, long func, int vector, int level, int prio );
#elif ( defined MCF5270 )
void SetIntc( long func, int vector, int level, int prio );
#else
#error Processor Not Defined
#endif
void UserMain(void * pd);
}

void PrintSRAMUsage();

/*----------------------------------------------------------------------------------*/
#ifdef RamTestEnable
#define SDRAMTestSize ( 256 * 1024  )
#define SRAMTestSize ( 4 * 1024  )
DWORD SDRAMTestBuff[SDRAMTestSize]__attribute__((aligned( 16 )));
DWORD SRAMTestBuff[SRAMTestSize]__attribute__((aligned( 16 ))) FAST_USER_VAR;
#endif

#ifdef OS_TEST
//
// Task Switch Variables
//
void MyTask( void *pdata );
double PendTest( DWORD Pends );
#define PEND_LIMIT (100000)
#define MY_TASK_STACK_SIZE (256)
DWORD MyTaskStack[MY_TASK_STACK_SIZE]__attribute__((aligned( 4 ))) FAST_USER_STK;
OS_SEM MySemaphore FAST_USER_VAR;
#endif

double TotalTime __attribute__((aligned( 16 ))) FAST_USER_VAR;

#ifdef IRQ_TEST
//
// IRQ Switch Variables
//
#define TOTAL_ISRS (1000000)
#define TOTAL_NM_ISRS (1000000)

#if (!defined MOD5272 || !defined SB72 || !defined SB72EX )
volatile DWORD ISRcount __attribute__((aligned( 4 ))) FAST_USER_VAR;
#endif
#endif

#ifdef UDP_TEST
//
// UDP Upload Variables
//
#define portnum (0x1234)
IPADDR ipaddr FAST_USER_VAR;
void SendPacket();
void UDPTest();
volatile int Pleft FAST_USER_VAR;
IPADDR UDPDestIP FAST_USER_VAR;
#define UDP_BUFFER_SIZE 1500
char UDPDataBuffer[UDP_BUFFER_SIZE]__attribute__((aligned( 4 ))) FAST_USER_VAR;
#endif

#ifdef TCP_TEST
//
// TCP Upload Variables
//
void TCPTest();
#define TCP_BUFFER_SIZE 20000
char TCPDataBuffer[TCP_BUFFER_SIZE]__attribute__((aligned( 16 )));// FAST_USER_VAR;
#endif


char input[16];

/*----------------------------------------------------------------------------------*/
 #ifdef IRQ_TEST

#if(!defined MOD5272 && !defined SB72 && !defined SB72EX && !defined MCF5208 && !defined MCF53017 )
INTERRUPT( IRQ5_ISR, 0x2600 )
{
	ISRcount ++;
	sim.eport.epfr = 0x20;   // Clear any pending irq flag
}
INTERRUPT( IRQ7_ISR, 0x2700 ){}

/*----------------------------------------------------------------------------------*/

void Function_Holder()
{
   __asm__  (" .global  NMI_ASM_Part"); // export the label for ASM part
   __asm__  (" .extern  NMI_C_Part");   // label for the C part of the ISR
   __asm__  ("NMI_ASM_Part:");          // label for the assembly part of the ISR
  // __asm__  (" move.w #0x2700,%sr ");   // set the IRQ mask to mask all
	__asm__  (" addq.l  #1, (ISRcount) "); //
	__asm__  (" move.l %d0, %a7@- "); //
	__asm__  (" move.l #0x80, %d0 "); //
	__asm__  (" move.b %d0, 0x40130006 "); //
	__asm__  (" move.l %a7@+, %d0 ");
   __asm__  (" rte");
}
#endif
#endif


/*----------------------------------------------------------------------------------*/
/************USER MAIN******************************************************/
/*----------------------------------------------------------------------------------*/

void UserMain(void * pd) {
    InitializeStack();
    if (EthernetIP == 0) GetDHCPAddress();
    OSChangePrio(MAIN_PRIO);
    EnableAutoUpdate();
    EnableTaskMonitor();

    #ifndef _DEBUG
    EnableSmartTraps();
    #endif

    #ifdef _DEBUG
    InitializeNetworkGDB_and_Wait();
    #endif

    printf("%s application started\r\n\r\n", AppName );

#ifdef UDP_TEST
    printf("Enter the IP for the destination of the UDP test: " );
    gets( UDPDataBuffer );
    UDPDestIP = AsciiToIp( UDPDataBuffer );
#endif

    for(int i =1; i< UDP_BUFFER_SIZE; i++)
    	    UDPDataBuffer[i-1] = i;

#ifdef SRAM_INFO
	 PrintSRAMUsage();
#endif

	// OPEN THE DUAL PORT SRAM
	// This can increase Ethernet throughput but also Task Switch and IRQ Switching Time
	//sim.scm.rambar = 0x20000200;
	//asm("movel #0x20000201, %d0");
	//asm("movec %d0, #0x0C05");

    StopWatchInit();

#ifdef RamTestEnable

    printf("\r\nStarting RAM speed test\r\n");

    StopWatchStart();
    for(volatile int i = 0; i<16; i++)
    	memset( SDRAMTestBuff, 0, (SDRAMTestSize*4) );
    TotalTime = StopWatchReadTime();
    StopWatchStopClear();
    printf("Memset clearing of %gMB of external SDRAM bandwidth %gMB/s\n\r", ((double)(SDRAMTestSize*4*16)/(double)(1024*1024)), ( ((double)(SDRAMTestSize*4*16)/(double)(1024*1024)) / (double)TotalTime) );

    StopWatchStart();
	for(volatile int i = 0; i<2048; i++)
		memset( SRAMTestBuff, 0, (SRAMTestSize*4) );
	TotalTime = StopWatchReadTime();
	StopWatchStopClear();
	printf("Memset clearing of %gMB of internal SRAM bandwidth %gMB/s\n\r", ((double)(SRAMTestSize*4*2048)/(double)(1024*1024)), ( ((double)(SRAMTestSize*4*2048)/(double)(1024*1024)) / (double)TotalTime) );

	// Put incrementing data in buffers
	for(volatile int i = 0; i<SDRAMTestSize; i++)
		SDRAMTestBuff[i] = i;
	for(volatile int i = 0; i<SRAMTestSize; i++)
		SDRAMTestBuff[i] = i;

    StopWatchStart();
    for(volatile int i = 0; i<16; i++)
		memcpy( (BYTE*)SDRAMTestBuff+(SDRAMTestSize*2), SDRAMTestBuff, (SDRAMTestSize*2) );
    TotalTime = StopWatchReadTime();
    StopWatchStopClear();
    printf("Moving %gMB from external SDRAM to external SDRAM bandwidth %gMB/s\n\r", ((double)(SDRAMTestSize*2*16)/(double)(1024*1024)), ( ((double)(SDRAMTestSize*2*16)/(double)(1024*1024)) / (double)TotalTime) );

    StopWatchStart();
    for(volatile int i = 0; i<1024; i++)
		memcpy( (BYTE*)SRAMTestBuff, SDRAMTestBuff, (SRAMTestSize*4) );
    TotalTime = StopWatchReadTime();
    StopWatchStopClear();
    printf("Moving %gMB from external SDRAM to internal SRAM bandwidth %gMB/s\n\r", ((double)(SRAMTestSize*4*1024)/(double)(1024*1024)), ( ((double)(SRAMTestSize*4*1024)/(double)(1024*1024)) / (double)TotalTime) );

    StopWatchStart();
    for(volatile int i = 0; i<1024; i++)
		memcpy( SDRAMTestBuff, (BYTE*)SRAMTestBuff, (SRAMTestSize*4) );
    TotalTime = StopWatchReadTime();
    StopWatchStopClear();
    printf("Moving %gMB from internal SRAM to external SDRAM bandwidth %gMB/s\n\r",  ((double)(SRAMTestSize*4*1024)/(double)(1024*1024)), ( ((double)(SRAMTestSize*4*1024)/(double)(1024*1024)) / (double)TotalTime) );

    StopWatchStart();
    for(volatile int i = 0; i<4096; i++)
		memcpy( (BYTE*)SRAMTestBuff+(SRAMTestSize*2), SRAMTestBuff, (SRAMTestSize*2) );
    TotalTime = StopWatchReadTime();
    StopWatchStopClear();
    printf("Moving %gMB from internal SRAM to internal SRAM bandwidth %gMB/s\n\r", ((double)(SRAMTestSize*2*4096)/(double)(1024*1024)), ( ((double)(SRAMTestSize*2*4096)/(double)(1024*1024)) / (double)TotalTime) );
#endif


#ifdef OS_TEST
    OSSemInit( &MySemaphore, 0 );
	 if ( OSTaskCreate( MyTask,
                      NULL,
                      ( void * ) &MyTaskStack[MY_TASK_STACK_SIZE],
                      ( void * ) MyTaskStack,
                      MAIN_PRIO+1 ) != OS_NO_ERR )
  	 {
      printf( "*** Error creating task\r\n" );
  	 }


    printf("\r\nStarting Task Switch Latency Test\r\n");
    PendTest( 1 );
    PendTest( PEND_LIMIT ); //about 40 seconds
    printf("%d Pends took %g Seconds or %g Seconds Per Pend\n\r", PEND_LIMIT, TotalTime, TotalTime/(double)PEND_LIMIT);
    printf("Each pend is 2 OS Task Switches which = %g seconds per switch\n\r\r\n", TotalTime/((double)PEND_LIMIT * (double)2));
#endif

 /*----------------------------------------------------------------------------------*/
 #ifdef IRQ_TEST

#if(!defined MOD5272 && !defined SB72 && !defined SB72EX && !defined MCF5208 && !defined MCF53017 )
    printf("Starting Standard IRQ Latency Test\r\n");
    sim.eport.epddr |= 0x20;  //sets IRQ5 in direction register to be an output
    sim.eport.epdr &= ~0x20;  //drive IRQ5 low
    sim.eport.eppar |= 0xC00;  //set to trigger on any edge
    sim.eport.epfr |= 0x20;   // Clear any pending irq flag
    sim.eport.epier |= 0x20;  //enable IRQ5 interrupts
    #if( defined MOD5282 || defined MOD5234 || defined CB34EX || defined MCF53017 )
    SetIntc( 0 , ( long )&IRQ5_ISR, 5, 5, 5 );   // initialize IRQ5 to use SRAM ISR
    #else
    SetIntc( ( long )&IRQ5_ISR, 5, 5, 5 );   // initialize IRQ5 to use SRAM ISR
    #endif
    ISRcount = 0;
    
    StopWatchStart();
    while( ISRcount < TOTAL_ISRS )
		sim.eport.epdr ^= 0x20;  //drive IRQ5 high
    TotalTime = StopWatchReadTime();
    StopWatchStopClear();
    sim.eport.epier &= ~0x20;  //disable IRQ5 interrupts
    printf("%d interrupts took %g Seconds or %g Seconds Per interrupt\n\r", TOTAL_ISRS, TotalTime, TotalTime/(double)TOTAL_ISRS);
    printf("Each IRQ has 2 context switches which = %g seconds per IRQ switch\n\r\r\n", TotalTime/((double)TOTAL_ISRS * (double)2));

/*----------------------------------------------------------------------------------*/
#if ( !defined SB70LC )
    printf("Starting Non-Maskable IRQ Latency Test\r\n");
    sim.eport.epddr |= 0x80;  //sets IRQ7 in direction register to be an output
    sim.eport.epdr &= ~0x80;  //drive IRQ7 low
    sim.eport.eppar |= 0xC000;  //set to trigger on any edge
    sim.eport.epfr |= 0x80;   // Clear any pending irq flag
    sim.eport.epier |= 0x80;  //enable IRQ7 interrupts
#if ( defined MCF5282 || defined MCF5234 || defined MCF53017 )
    SetIntc( 0 , ( long )&NMI_ASM_Part, 7, 7, 7 );   // initialize IRQ7 to use SRAM ISR
#elif ( defined MCF5270 )
    SetIntc( ( long )&NMI_ASM_Part, 7, 7, 7 );   // initialize IRQ7 to use SRAM ISR
#else
#error PROCESSOR NOT DEFINED
    #endif
    ISRcount = 0;
    
    StopWatchStart();
    while( ISRcount <  TOTAL_NM_ISRS )
	 	sim.eport.epdr ^= 0x80;
    TotalTime = StopWatchReadTime();
    StopWatchStopClear();
    sim.eport.epier &= ~0x80;  //disable IRQ5 interrupts
    printf("%d interrupts took %g Seconds or %g Seconds Per Non-Maskable Interrupt\n\r", TOTAL_ISRS, TotalTime, TotalTime/(double)TOTAL_ISRS);
    printf("Each IRQ has 2 context switches which = %g seconds per Non-Maskable IRQ switch\n\r\r\n", TotalTime/((double)TOTAL_ISRS * (double)2));
#endif
#endif
#endif
/*----------------------------------------------------------------------------------*/

#ifdef UDP_TEST
	 UDPTest();
#endif

/*----------------------------------------------------------------------------------*/

#ifdef TCP_TEST
 	printf("Starting TCP Upload Data Payload Test\r\n");
	printf("Type \"speed <moduleIP>\" at the PC command prompt to start\r\n");
    register DWORD nsent=0;
    register int rv;
    for (int i=0; i< TCP_BUFFER_SIZE; i++) TCPDataBuffer[i]=('A'+ (i % 64));
    int fdListen=listen(INADDR_ANY,1234,5);
	if (fdListen>0)
	{
	 printf("FD listen is ok\r\n");
	 while(1)
	  {
#ifdef UDP_TEST
	  if(charavail())
	  {
		  gets(input);
		  UDPTest();
	  }
#endif

	  int fda= accept(fdListen,NULL,NULL,5);
	  if (fda>1)
	    {
	     printf("Accepted Waiting\r\n");
	    //setsockoption(fda,SO_NOPUSH);
	    //setsocketackbuffers(fda,40);
	     read(fda,TCPDataBuffer,1);
		  nsent=0;
		  StopWatchStopClear();
	     StopWatchStart();
	     while(1)
	      {
			 rv=write(fda,TCPDataBuffer,TCP_BUFFER_SEGMENTS * 1460);//BUFFER_SIZE);

	       if (rv<0)
	        {
	    	   TotalTime = StopWatchReadTime();
			 close(fda);
			 break;
		   }
		   else
			nsent+=rv;
	      }

    		StopWatchStopClear();
    		printf("%gMByte of TCP data upload took %g seconds or %gMByte/s\n\r",
    		(double)(nsent)/(double)(1024*1024), TotalTime,
    		(double)(1/TotalTime) * (double)((double)(nsent)/(double)(1024*1024)));
    	  	printf("TCP Data Payload = %g Mbps\n\r\r\n", ( 1.0 / TotalTime ) * (double)((double)(nsent)/(double)(1024*1024)) * 8.0  );

	    }
	    else if (fda == 0)
	    printf("Accepfailed\r\n");
	  }
	 }
	else
	  printf("FD listen is NOT ok\r\n");

#ifdef UDP_TEST
	printf("Press any key to go again\r\n");
	gets(input);
#endif
#endif
}


#ifdef OS_TEST
/*-------------------------------------------------------------------
 MyTask
 This task will post to a semaphore
 -------------------------------------------------------------------*/
void MyTask( void *pdata )
{
	while(1)
		OSSemPost( &MySemaphore );
}
double PendTest( DWORD Pends )
{
   StopWatchStart();
   while( Pends )
   {
      OSSemPend( &MySemaphore, 0 );
      Pends--;
	}
   TotalTime = StopWatchReadTime();
   StopWatchStopClear();
   return TotalTime;
}
#endif



#ifdef UDP_TEST

void UDPTest()

{
	#ifndef _DEBUG
	 while( !EtherLink());
    #else
    printf("We are in debug mode\r\n");
    #endif
    {
       printf("Starting UDP Upload Throughput Test\r\n");
#ifdef UDP_SEND_SOCKET
       int fdUDP = CreateTxUdpSocket(UDPDestIP, portnum, portnum);
       //int fdUDP = CreateRxTxUdpSocket(UDPDestIP, portnum, portnum);
#endif
       //Packet = 1500 Bytes, 10MBytes = 10485760 Bytes = 6990.5 packets
       StopWatchStart();
       #define TotalPackets (6990)
#ifdef UDP_SEND_KEEP
      {
      UDPPacket pkt;
		pkt.SetSourcePort( portnum );
		pkt.SetDestinationPort( portnum );
		pkt.AddData( (BYTE*)UDPDataBuffer, MAX_UDPDATA );
		for( Pleft = 0; Pleft < (TotalPackets-1); Pleft++ )
			pkt.SendAndKeep( UDPDestIP );

		pkt.Send( UDPDestIP );
      }
#elif defined UDP_SEND_NEW
       for( Pleft = 0; Pleft < TotalPackets; Pleft++ )
       {
    	UDPPacket pkt;
		pkt.SetSourcePort( portnum );
		pkt.SetDestinationPort( portnum );
		pkt.AddData( (BYTE*)UDPDataBuffer, (MAX_UDPDATA) );
		pkt.Send( UDPDestIP );
      }
#elif defined UDP_SEND_SOCKET
       {
    	for( Pleft = 0; Pleft < TotalPackets; Pleft++ )
    		writeall(fdUDP, UDPDataBuffer, MAX_UDPDATA );
      }
#endif
    }
    TotalTime = StopWatchReadTime();
    StopWatchStopClear();
    printf("10MByte UDP upload took %g seconds\r\n", TotalTime );
  	printf("Total UDP Throughput = %g Mbps\n\r", ( 1.0 / TotalTime ) * ( ( TotalPackets * 1500.0 * 8.0 ) / 1000000.0 ) );
  	printf("UDP Data Payload = %g Mbps\n\r\r\n", ( 1.0 / TotalTime ) * ( ( MAX_UDPDATA * TotalPackets * 8.0 ) / 1000000.0 ) );
}
#endif

#ifdef SRAM_INFO
void PrintSRAMUsage( )
{
#ifdef ENABLE_SRAM_SYS
	 extern BYTE SRAM_START, SRAM_END,
     				 SYS_VAR_SECT_END, USER_VAR_SECT_END, IDLE_STK_SECT_END, MAIN_STK_SECT_END,
    				 ETHER_STK_SECT_END, IP_STK_SECT_END, TCP_STK_SECT_END, HTTP_STK_SECT_END,
    				 FTP_STK_SECT_END, WIFI_STK_SECT_END, PPP_STK_SECT_END, COMMAND_STK_SECT_END,
    				 USER_STK_SECT_END, BUFFERS_VAR_SECT_END, ETHER_VAR_SECT_END, IP_VAR_SECT_END,
    				 TCP_VAR_SECT_END;
    printf("\r\n\r\nYou are using %u bytes in SRAM\r\n", (unsigned int)&SRAM_END - (unsigned int)&SRAM_START);
    printf("SYS_VAR_SECT = %u bytes\r\n", 	(unsigned int)&SYS_VAR_SECT_END - (unsigned int)&SRAM_START);
    printf("IDLE_STK_SECT = %u bytes\r\n", 	(unsigned int)&IDLE_STK_SECT_END - (unsigned int)&SYS_VAR_SECT_END);
	printf("MAIN_STK_SECT = %u bytes\r\n", 	(unsigned int)&MAIN_STK_SECT_END - (unsigned int)&IDLE_STK_SECT_END);
	printf("ETHER_VAR_SECT = %u bytes\r\n", 	(unsigned int)&ETHER_VAR_SECT_END - (unsigned int)&MAIN_STK_SECT_END);
	printf("ETHER_STK_SECT = %u bytes\r\n", 	(unsigned int)&ETHER_STK_SECT_END - (unsigned int)&ETHER_VAR_SECT_END);
	printf("IP_VAR_SECT = %u bytes\r\n", 		(unsigned int)&IP_VAR_SECT_END - (unsigned int)&ETHER_STK_SECT_END);
	printf("IP_STK_SECT = %u bytes\r\n", 		(unsigned int)&IP_STK_SECT_END - (unsigned int)&IP_VAR_SECT_END);
	printf("TCP_VAR_SECT = %u bytes\r\n", 	(unsigned int)&TCP_VAR_SECT_END - (unsigned int)&IP_STK_SECT_END);
	printf("TCP_STK_SECT = %u bytes\r\n", 	(unsigned int)&TCP_STK_SECT_END - (unsigned int)&TCP_VAR_SECT_END);
	printf("HTTP_STK_SECT = %u bytes\r\n", 	(unsigned int)&HTTP_STK_SECT_END - (unsigned int)&TCP_STK_SECT_END);
    printf("FTP_STK_SECT = %u bytes\r\n",  	(unsigned int)&FTP_STK_SECT_END - (unsigned int)&HTTP_STK_SECT_END);
	printf("WIFI_STK_SECT = %u bytes\r\n", 	(unsigned int)&WIFI_STK_SECT_END - (unsigned int)&FTP_STK_SECT_END);
	printf("PPP_STK_SECT = %u bytes\r\n", 	(unsigned int)&PPP_STK_SECT_END - (unsigned int)&WIFI_STK_SECT_END);
	printf("COMMAND_STK_SECT = %u bytes\r\n",(unsigned int)&COMMAND_STK_SECT_END - (unsigned int)&PPP_STK_SECT_END);
	printf("USER_VAR_SECT = %u bytes\r\n", 	(unsigned int)&USER_VAR_SECT_END - (unsigned int)&COMMAND_STK_SECT_END);
	printf("USER_STK_SECT = %u bytes\r\n", 	(unsigned int)&USER_STK_SECT_END - (unsigned int)&USER_VAR_SECT_END);
	printf("BUFF_VAR_SECT = %u bytes\r\n", 	(unsigned int)&BUFFERS_VAR_SECT_END - (unsigned int)&USER_STK_SECT_END);
#ifdef FAST_BUFFERS
    extern unsigned int FAST_BUFFER_POOL_SIZE;
	printf("There are %u fast buffers in the SRAM pool\r\n", FAST_BUFFER_POOL_SIZE);
	extern buffer_list free_fast_list;
	printf("There are %u fast buffers free (available)\r\n", free_fast_list.GetCount());
#endif
#endif
}
#endif
